"""
:class:`Input4MIPsDataset` and associated metadata
"""
from __future__ import annotations

import datetime as dt
import os.path
import uuid
from functools import partial
from typing import TYPE_CHECKING, Any

import cftime
import xarray as xr
from attrs import asdict, define, field
from attrs.validators import in_, matches_re

from carpet_concentrations.input4MIPs.metadata_options import (
    ACTIVITY_ID_OPTIONS,
    CONVENTION_OPTIONS,
    CREATION_DATE_REGEX,
    DATASET_CATEGORY_OPTIONS,
    INCLUDES_EMAIL_REGEX,
    UUID_REGEX,
)
from carpet_concentrations.time import get_start_of_next_month, split_time_to_year_month

if TYPE_CHECKING:
    from pathlib import Path


# If you're thinking about sub-classing this to update it for e.g. CMIP7,
# please consider instead refactoring to use the builder pattern. That will
# make the business logic and creation choices easier to follow for future
# developers (and the business logic really belongs to the class creation,
# once the rules about what can go in the class are decided, everything else
# follows pretty simply).
@define
class Input4MIPsMetadata:
    """
    Input4MIPs metadata

    These are all required fields.

    Notes
    -----
    `variable_id` is not included here because it should be derived from the
    data (which is combined with the metadata elsewhere).
    """

    activity_id: str = field(validator=in_(ACTIVITY_ID_OPTIONS))
    """Activity ID of the dataset"""

    contact: str = field(validator=matches_re(INCLUDES_EMAIL_REGEX))
    """Contact for the dataset"""

    Conventions: str = field(validator=in_(CONVENTION_OPTIONS))
    """CF conventions adhered to by the dataset"""

    dataset_category: str = field(validator=in_(DATASET_CATEGORY_OPTIONS))
    """Datset category"""

    frequency: str
    """Time frequency of the dataset"""

    further_info_url: str
    """URL with further information"""

    grid_label: str
    """Grid label of the dataset"""

    institution: str
    """Institution that produced the dataset"""

    institution_id: str
    """Unique ID of the institution that produced the dataset"""

    mip_era: str
    """MIP era of the dataset"""

    nominal_resolution: str
    """Nominal resolution of the dataset"""

    realm: str
    """Realm of the dataset"""

    source_version: str
    """Version of the dataset"""

    source_id: str
    """Source id of the dataset"""

    source: str
    """Source of the dataset (human-readable)"""

    target_mip: str
    """Target MIP of the dataset"""

    title: str
    """Title of the dataset (human-readable)"""

    def to_dataset_attributes(self) -> dict[str, str]:
        """
        Convert to a format that can be used as dataset attributes
        """
        out = {k: v for k, v in asdict(self).items()}

        return out


# As above, if you're thinking about sub-classing this to update it for e.g.
# CMIP7, please consider instead refactoring to use the builder pattern for
# the same reasons as above.
@define
class Input4MIPsMetadataOptional:
    """
    Input4MIPs optional metadata

    These are all optional fields.

    Notes
    -----
    This is currently written such that no fields outside of these can be
    provided. We don't fully understand the input4MIPs rules, so this could
    easily be the wrong choice. Refactoring should be relatively
    straightforward if needed. It would make sense that these fields are
    locked to avoid clashes with compulsory metadata...?
    """

    comment: str | None = None
    """Comment on the dataset"""

    # No idea if this can be validated or not nor what dataspecs are being
    # referred to
    data_specs_version: str | None = None
    """Data specs version used when creating the dataset"""

    external_variables: str | None = None  # No idea of comma separated or what
    """
    Variables relevant to the dataset that aren't included in the dataset itself

    For example, cell area variables like 'areacella'
    """

    grid: str | None = None  # No idea if this is meant to follow a controlled vocab
    """Human-readable version of the grid on which the dataset applies"""

    history: str | None = None
    """File modification history"""

    product: str | None = None  # No idea if this is meant to follow a controlled vocab
    """Product the data represents"""

    references: str | None = None
    """References related to the dataset"""

    region: str | None = None  # No idea if this is meant to follow a controlled vocab
    """Region to which the dataset applies"""

    release_year: str | None = None  # TODO: add validation that it is a year
    """Release year of the dataset"""

    source_description: str | None = (
        None  # No idea if this is meant to follow a controlled vocab
    )
    """Description of the dataset's source"""

    source_type: str | None = (
        None  # No idea if this is meant to follow a controlled vocab
    )
    """Description of the type of the dataset's source"""

    table_id: str | None = None  # No idea if this is meant to follow a controlled vocab
    """No idea, maybe the CMOR table used to write the dataset"""

    table_info: str | None = (
        None  # No idea if this is meant to follow a controlled vocab
    )
    """No idea, maybe info about the CMOR table used to write the dataset"""

    license: str | None = None
    """License information"""

    def to_dataset_attributes(self) -> dict[str, str]:
        """
        Convert to a format that can be used as dataset attributes
        """
        out = {k: v for k, v in asdict(self).items() if v is not None}

        return out


# As above, if you're thinking about sub-classing this to update it for e.g.
# CMIP7, please consider instead refactoring to use the builder pattern for
# the same reasons as above.
@define
class Input4MIPsDataset:
    """
    Input4MIPs dataset

    Holds input4MIPs data and also helps write them to disk in a way that
    conforms to input4MIPs standards
    """

    ds: xr.Dataset
    """
    Dataset
    """
    # Checks to add (and test):
    # - variable_id in attributes matches data (hence only one variable)
    # - variable dimensions matches expected dimensions
    # - all dimensions variables have bounds
    # - source_version and source_id are consistent
    # - assert metadata consistent with Input4MIPsMetadata
    # - grid label matches data
    # - nominal_resolution matches data
    # - realm matches data
    # - source id matches version and institution
    # - variable_id matches data
    # - check dataset_category against data
    # - check frequency against data
    # - data pint quantified so we can use cf-xarray to then go to UDUNITS
    # - all compulsory metadata fields are in ds.attrs (can use fields(
    #   Metadata))
    # - no forbidden metadata fields are in ds.attrs (would require defining
    #   forbidden fields first)

    directory_template: str = os.path.join(
        "{activity_id}",
        "{mip_era}",
        "{target_mip}",
        "{institution_id}",
        "{source_id}",
        "{realm}",
        "{frequency}",
        "{variable_id}",
        "{grid_label}",
        "v{version}",
    )
    """
    Template used to determine the directory in which to save the data
    """

    filename_template: str = "_".join(
        [
            "{variable_id}",
            "{activity_id}",
            "{dataset_category}",
            "{target_mip}",
            "{source_id}",
            "{grid_label}",
            "{start_date}",
            "{end_date}.nc",
        ]
    )
    """
    Template used to determine the filename when saving the data
    """

    @classmethod
    def from_metadata_autoadd_bounds_to_dimensions(  # noqa: PLR0913
        cls,
        ds: xr.Dataset,
        dimensions: tuple[str, ...],
        metadata: Input4MIPsMetadata,
        metadata_optional: Input4MIPsMetadataOptional | None = None,
        time_dimension: str = "time",
        monthly_time_bounds: bool = True,
        copy: bool = True,
        **kwargs: Any,
    ) -> Input4MIPsDataset:
        """
        Create instance from metadata and an unbounded dataset

        For the given dimensions, bounds are checked and added if needed. The
        metadata is then used to fill out ``ds``'s metadata before
        initialising.

        Parameters
        ----------
        ds
            Dataset

        dimensions
            Dimensions of the dataset, these are checked for appropriate
            bounds.

        metadata
            Metadata (required)

        metadata_optional
            Optional metadata

        time_dimension
            The name of the time dimension. This is provided to give full
            control of the application of ``monthly_time_bounds`` to the user.

        monthly_time_bounds
            Should added time bounds cover each month? This is needed for data
            on a monthly timestep because the middle of each timestep is not
            the start and end of the month in the case when subsequent months
            don't have the same number of days.

        copy
            Should a copy of the dataset be made? If no, the data is modified
            in place which can cause unexpected changes if references are not
            appropriately managed.

        **kwargs
            Other initialisation arguments for the instance. They are passed
            directly to the constructor.

        Returns
        -------
            Prepared instance

        Raises
        ------
        AssertionError
            ``ds.attrs`` is already set or there is more than one variable in ``ds``
        """
        if copy:
            ds = ds.copy(deep=True)
        else:
            raise NotImplementedError(copy)

        if ds.attrs:
            raise AssertionError("All metadata should be autogenerated")  # noqa: TRY003

        if len(ds.data_vars) == 1:
            variable_id = list(ds.data_vars.keys())[0]
        else:
            raise AssertionError("Can only write one variable per file")  # noqa: TRY003

        # add extra metadata following CF conventions, not really sure what
        # this does but it's free so we include it on the assumption that they
        # know more than we do (may be a bad assumption of course...)
        ds = ds.cf.guess_coord_axis().cf.add_canonical_attributes()

        # add bounds to dimensions
        for dim in dimensions:
            if dim == time_dimension:
                ds = add_time_bounds(ds, monthly_time_bounds)
            else:
                ds = ds.cf.add_bounds(dim)

        # transpose to match dimensions
        ds = ds.transpose(*dimensions, ...)

        # Get info from metadata
        attributes = {"variable_id": variable_id, **metadata.to_dataset_attributes()}
        if metadata_optional is not None:
            attributes.update(metadata_optional.to_dataset_attributes())

        ds.attrs = attributes

        return cls(ds, **kwargs)

    def write(
        self,
        root_data_dir: Path,
        unlimited_dims: tuple[str, ...] = ("time",),
        encoding_kwargs: dict[str, Any] | None = None,
    ) -> Path:
        """
        Write to disk

        Parameters
        ----------
        root_data_dir
            Root directory in which to write the file

        unlimited_dims
            Dimensions which should be unlimited

        encoding_kwargs
            Kwargs to use when encoding to disk. These are passed to
            :meth:`xr.Dataset.to_netcdf`

        Returns
        -------
            Where the file was written
        """
        if encoding_kwargs is None:
            encoding_kwargs = {"zlib": True, "complevel": 5}

        # Can shallow copy here as we don't need to worry about mangling the
        # data as the ref is not retured
        ds_disk = self.ds.copy(deep=False).pint.dequantify(format="cf")

        # Unique for every written file, so we don't provide a way for the
        # user to overwrite this at present
        ds_disk.attrs["tracking_id"] = generate_tracking_id()
        ds_disk.attrs["creation_date"] = generate_creation_timestamp()

        verify_disk_ready(ds_disk)

        out_path = self.get_filepath(
            ds_disk,
            root_data_dir,
        )

        out_path.parent.mkdir(parents=True, exist_ok=True)

        ds_disk.to_netcdf(
            out_path,
            unlimited_dims=unlimited_dims,
            encoding={ds_disk.attrs["variable_id"]: encoding_kwargs},
        )

        return out_path

    def get_filepath(
        self,
        ds_disk: xr.Dataset,
        root_data_dir: Path,
    ) -> Path:
        """
        Get filepath

        Parameters
        ----------
        ds_disk
            Disk ready dataset

        root_data_dir
            Root directory in which to generate the filepath

        Returns
        -------
            Filepath
        """
        format_date_h = partial(format_date, ds_frequency=ds_disk.attrs["frequency"])
        avail_metadata = {
            **ds_disk.attrs,
            "version": get_version(ds_disk.attrs["creation_date"]),
            "start_date": format_date_h(ds_disk.time.values.min()),
            "end_date": format_date_h(ds_disk.time.values.max()),
        }
        # This will likely require refactoring to become injectable
        avail_metadata_file_compat = {
            k: v.replace("_", "-") for k, v in avail_metadata.items()
        }

        out_dir = self.directory_template.format(**avail_metadata_file_compat)
        out_fname = self.filename_template.format(**avail_metadata_file_compat)

        return root_data_dir / out_dir / out_fname


def format_date(
    date: cftime.datetime | dt.datetime,
    ds_frequency: str,
) -> str:
    """
    Format date for filepath

    Parameters
    ----------
    date
        Date to format

    ds_frequency
        Frequency of the underlying dataset

    Returns
    -------
        Formatted date
    """
    if ds_frequency.endswith("mon"):
        datestring = date.strftime("%Y%m")
    else:
        raise NotImplementedError(ds_frequency)

    return datestring


def get_version(creation_date: str) -> str:
    """
    Get version string for filepath

    Parameters
    ----------
    creation_date
        Creation date

    Returns
    -------
        Version string
    """
    return dt.datetime.strptime(creation_date, "%Y-%m-%dT%H:%M:%SZ").strftime("%Y%m%d")


def add_time_bounds(
    ds: xr.Dataset,
    monthly_time_bounds: bool = False,
    output_dim: str = "bounds",
) -> xr.Dataset:
    """
    Add time bounds to a dataset

    This should be pushed upstream to cf-xarray at some point probably

    Parameters
    ----------
    ds
        Dataset to which to add time bounds

    monthly_time_bounds
        Are we looking at monthly data i.e. should the time bounds run from
        the start of one month to the next (which isn't regular spacing but is
        most often what is desired/required)

    Returns
    -------
        Dataset with time bounds

    Notes
    -----
    There is no copy here, ``ds``is modified in place (copy before passing if
    you don't want this)
    """
    # based on cf-xarray's implementation, to be pushed back upstream at some
    # point
    # Step 1 of push back upstream:
    variable = "time"
    bname = f"{variable}_bounds"

    if bname in ds.variables:
        raise ValueError(  # noqa: TRY003
            f"Bounds variable name {bname!r} will conflict!"
        )

    if monthly_time_bounds:
        ds_ym = split_time_to_year_month(ds, time_axis=variable)

        # This may need to be refactored to allow the cftime_converter to be
        # injected, same idea as `convert_to_time`
        bounds = xr.DataArray(
            [
                [cftime.datetime(y, m, 1), get_start_of_next_month(y, m)]
                for y, m in zip(ds_ym.year, ds_ym.month)
            ],
            dims=(variable, "bounds"),
            coords={variable: ds[variable], "bounds": [0, 1]},
        ).transpose(..., "bounds")
    else:
        # This will require some thinking because `ds.cf.add_bounds(dim)`
        # doesn't work with cftime.datetime objects. Probably needs an issue upstream
        # and then a monkey patch or custom function here as a workaround.
        raise NotImplementedError(monthly_time_bounds)

    ds.coords[bname] = bounds
    ds[variable].attrs["bounds"] = bname

    return ds


def verify_disk_ready(ds: xr.Dataset) -> None:
    """
    Verify that a dataset is disk ready

    Parameters
    ----------
    ds
        Dataset to check

    Notes
    -----
    Very rough, doesn't really do anything right now
    """
    # call verify as a final check before writing
    # Note that we could change write to so it wraps around
    # [CMOR](https://cmor.llnl.gov/)
    # I'm not sure a) how often CMOR is used or b) how helpful it is
    # compared to just writing the same functionality here. We will have
    # to have a play and also ask CMOR devs what their roadmap looks like
    # I assume. I seem to have thought it looked promising here
    # https://github.com/PCMDI/cmor3_documentation/pull/57/files
    # This might also be worth looking at because it claims to implement CF
    # conventions: https://ncas-cms.github.io/cf-python/
    if not CREATION_DATE_REGEX.fullmatch(ds.attrs["creation_date"]):
        raise AssertionError(  # noqa: TRY003
            f"creation_date must match {CREATION_DATE_REGEX!r}"
        )

    if not UUID_REGEX.fullmatch(ds.attrs["tracking_id"]):
        raise AssertionError(f"tracking_id must match {UUID_REGEX}")  # noqa: TRY003

    if not ds["time"].encoding:
        raise AssertionError(  # noqa: TRY003
            "Not specifying a time encoding will cause all sorts of headaches"
        )


def generate_tracking_id() -> str:
    """
    Generate tracking ID

    Returns
    -------
        Tracking ID
    """
    return "hdl:21.14100/" + str(uuid.uuid4())


def generate_creation_timestamp() -> str:
    """
    Generate creation timestamp, formatted as needed for input4MIPs files

    Returns
    -------
        Creation timestamp
    """
    ts = dt.datetime.utcnow().replace(
        microsecond=0  # remove microseconds from creation_timestamp
    )

    return f"{ts.isoformat()}Z"  # Z indicates timezone is UTC



"""
Tools for helping with :mod:`attrs`, particularly validators
"""
from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, TypeVar

    import attr

    T = TypeVar("T")


def add_attrs_context(
    original: Callable[[Any, attr.Attribute[Any], T], None]
) -> Callable[[Any, attr.Attribute[Any], T], None]:
    """
    Decorate function with a ``try...except`` to add the :mod:`attrs` context

    This means that the information about what attribute was being set and
    what value it was passed is also shown to the user

    Parameters
    ----------
    original
        Function to decorate

    Returns
    -------
        Decorated function

    Notes
    -----
    Only works with Python 3.11 and above. For other Python versions, the raw
    error is simply shown instead
    """

    @wraps(original)
    def with_attrs_context(
        instance: Any,
        attribute: attr.Attribute[Any],
        value: T,
    ) -> None:
        try:
            original(instance, attribute, value)
        except Exception as exc:
            if hasattr(exc, "add_note"):
                exc.add_note(
                    "\nError raised while initialising attribute "
                    f"``{attribute.name}`` of ``{type(instance)}``. "
                    f"\nValue provided: {value}"
                )

            raise

    return with_attrs_context


def make_attrs_validator_compatible_single_input(
    func_to_wrap: Callable[[T], None],
) -> Callable[[Any, attr.Attribute[Any], T], None]:
    """
    Create a function that is compatible with validation via :func:`attrs.field`

    This assumes that the function you're wrapping only takes a single input.

    Parameters
    ----------
    func_to_wrap
        Function to wrap

    Returns
    -------
        Wrapped function, which can be used as a validator with
        :func:`attrs.field`
    """

    @add_attrs_context
    @wraps(func_to_wrap)
    def attrs_compatible(
        instance: Any,
        attribute: attr.Attribute[Any],
        value: T,
    ) -> None:
        func_to_wrap(value)

    return attrs_compatible


def make_attrs_validator_compatible_value_instance_input(
    func_to_wrap: Callable[[Any, T], None],
) -> Callable[[Any, attr.Attribute[Any], T], None]:
    """
    Create a function that is compatible with validation via :func:`attrs.field`

    This assumes that the function you're wrapping takes the instance and the
    values as inputs.

    Parameters
    ----------
    func_to_wrap
        Function to wrap

    Returns
    -------
        Wrapped function, which can be used as a validator with
        :func:`attrs.field`
    """

    @add_attrs_context
    @wraps(func_to_wrap)
    def attrs_compatible(
        instance: Any,
        attribute: attr.Attribute[Any],
        value: T,
    ) -> None:
        func_to_wrap(instance, value)

    return attrs_compatible



"""
Gridder based on pre-calculated latitudinal gradient and seasonality
"""
from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any

import numpy as np
from attrs import define, field

from carpet_concentrations.attrs_utils import (
    make_attrs_validator_compatible_value_instance_input,
)
from carpet_concentrations.xarray_pint_utils import (
    check_pint_quantified_dataset,
    check_pint_quantified_dataset_attrs,
)
from carpet_concentrations.xarray_utils import (
    calculate_weighted_area_mean_latitude_only,
    check_all_units_compatible_attrs,
    check_dimensions,
)

if TYPE_CHECKING:
    import xarray as xr


def _attribute_has_year_month_lat_coords(
    instance: Any,
    value: xr.Dataset,
    attr_to_get: str,
) -> None:
    check_dimensions(
        value[getattr(instance, attr_to_get)], ("year", "month", "lat"), extras_ok=True
    )


_seasonality_has_year_month_lat_coords = (
    make_attrs_validator_compatible_value_instance_input(
        partial(_attribute_has_year_month_lat_coords, attr_to_get="seasonality_name")
    )
)


_latitudinal_gradient_has_year_month_lat_coords = (
    make_attrs_validator_compatible_value_instance_input(
        partial(
            _attribute_has_year_month_lat_coords,
            attr_to_get="latitudinal_gradient_name",
        )
    )
)


def _seasonality_annual_mean_zero(
    instance: Any,
    value: xr.Dataset,
) -> None:
    seasonality = value[instance.seasonality_name]
    np.testing.assert_allclose(
        seasonality.mean("month").data.magnitude,
        0,
        atol=1e-8,
        err_msg="seasonality must have an annual-mean of zero in all years",
    )


_seasonality_annual_mean_zero_attrs = (
    make_attrs_validator_compatible_value_instance_input(_seasonality_annual_mean_zero)
)


def _latitudinal_gradient_spatial_mean(
    instance: Any,
    value: xr.Dataset,
) -> None:
    np.testing.assert_allclose(
        calculate_weighted_area_mean_latitude_only(
            value, [instance.latitudinal_gradient_name]
        )[instance.latitudinal_gradient_name].data.magnitude,
        0,
        atol=1e-8,
        err_msg=(
            "latitudinal gradient must have an area-weighted spatial-mean "
            "of zero in all timesteps"
        ),
    )


_latitudinal_gradient_spatial_mean_zero_attrs = (
    make_attrs_validator_compatible_value_instance_input(
        _latitudinal_gradient_spatial_mean
    )
)


@define
class LatitudeSeasonalityGridder:
    """
    Gridder based on specified latitudinal gradient and seasonality
    """

    gridding_values: xr.Dataset = field(
        validator=[
            check_pint_quantified_dataset_attrs,
            _seasonality_has_year_month_lat_coords,
            _seasonality_annual_mean_zero_attrs,
            _latitudinal_gradient_has_year_month_lat_coords,
            _latitudinal_gradient_spatial_mean_zero_attrs,
            check_all_units_compatible_attrs,
        ]
    )
    """
    Values to use to convert global-mean concentrations to a grid

    This should have already been turned into a pint-compatiable quantity using
    ``pint.quantify`` or similar. It should also contain variables that match
    the value of ``seasonality_name`` and ``latitudinal_gradient_name``. Both
    the variables must have at least the dimensions ("year", "month", "lat").
    The seasonality variable must have an annual-mean of zero. The latitudinal
    gradient must have a spatial-mean of zero.
    """

    seasonality_name: str = "seasonality"
    """
    Name of the seasonality variable (in ``gridding_values`` but also used
    elsewhere)
    """

    latitudinal_gradient_name: str = "latitudinal_gradient"
    """
    Name of the latitudinal gradient variable (in ``gridding_values`` but also
    used elsewhere)
    """

    def calculate(self, global_means: xr.Dataset) -> xr.Dataset:
        r"""
        Calculate gridded values

        Parameters
        ----------
        global_means
            Global-mean values. These should have already been interpolated
            such that their dimensions are ``("year", "month")`` (other
            dimensions e.g. ``"scenario"`` are also ok) to obtain sensible
            results (if not, there may be jumps at the start and end of years).
            ``global_means`` should have also already been turned into a
            pint-compatiable quantity using ``pint.quantify`` or similar.

        Returns
        -------
            Gridded values on a ``("latitude", "year", "month")`` grid (plus
            any other dimensions present in ``global_means``)

        Raises
        ------
        CoordinateError
            ``global_means`` does not have at least the dimensions of
            ``("year", "month")`` (other dimensions e.g. ``"scenario"`` are
            also ok)

        NotPintQuantifiedError
            ``global_means`` has not been quantified using ``pint.quantify`` or
            similar

        Notes
        -----
        Eq. 1 of :footcite:t:`meinshausen_et_al_2017_concentrations`. We use
        slightly different notation here for clarity.

        .. math::

            C(l, m, y, ...) = \overline{C(m, y, ...)} + S(l, m, y) + L(l, m, y)

        i.e. the concentration at latitude :math:`l` in year :math:`y` and
        month :math:`m` is the sum of the global-, annual-mean concentration in
        that year (already interpolated down to a monthly timestep), the
        seasonality at that latitude in that year in that month (varies with
        year and latitude as often scaled with something else) and the
        latitudinal gradient at that latitude in that year and month (varies
        with year and month as often scaled with something else, must
        have monthly information to avoid step changes in output). Note that
        the global-, annual-mean concentrations :math:`C(y, m)` must be
        pre-interpolated onto monthly steps using a mean-preserving alogrithm
        [#11 for function which will do this] to avoid spurious steps
        in the outputs.

        The ellipses represent extra dimensions (e.g. scenario, greenhouse
        gas) that might be in :math:`C(y, m)`. Any such dimensions are
        preserved in the output.
        """
        check_pint_quantified_dataset(global_means)
        for darray in global_means.values():
            check_dimensions(darray, ("year", "month"), extras_ok=True)

        res = (
            global_means
            + self.gridding_values["seasonality"]
            + self.gridding_values["latitudinal_gradient"]
        )

        return res



"""
Time handling
"""
from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Protocol

import cftime
import numpy as np

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any

    import xarray as xr


MONTHS_PER_YEAR: int = 12
"""Months per year"""


class NonUniqueYearMonths(ValueError):
    """
    Raised when the user tries to convert to year-month with non-unique values

    This happens when the datetime values lead to year-month values that are
    not unique
    """

    def __init__(
        self, unique_vals: Iterable[tuple[int, int]], counts: Iterable[int]
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        unique_vals
            Unique values. In each tuple, the first value is the year and the
            second is the month.

        counts
            Counts of the number of time each unique value appeared in the
            original array
        """
        non_unique = list((v, c) for v, c in zip(unique_vals, counts) if c > 1)

        error_msg = (
            "Your year-month axis is not unique. "
            f"Year-month values with a count > 1: {non_unique}"
        )
        super().__init__(error_msg)


def convert_year_month_to_time(
    inp: xr.Dataset,
    day: int = 1,
    **kwargs: Any,
) -> xr.Dataset:
    """
    Convert year and month co-ordinates into a time axis

    This is a facade to :func:`convert_to_time`

    Parameters
    ----------
    inp
        Data to convert

    day
        Day of the month to assume in output

    **kwargs
        Passed to intialiser of :class:`cftime.datetime`

    Returns
    -------
        Data with time axis
    """
    return convert_to_time(
        inp,
        time_coords=("year", "month"),
        cftime_converter=partial(cftime.datetime, day=day, **kwargs),
    )


class CftimeConverter(Protocol):  # pylint: disable=too-few-public-methods
    """
    Callable that supports converting stacked time co-ordinates to :obj:`cftime.datetime`
    """

    def __call__(
        self,
        *args: np.float_ | np.int_,
    ) -> cftime.datetime:
        """
        Convert input values to an :obj:`cftime.datetime`
        """


def convert_to_time(
    inp: xr.Dataset,
    time_coords: tuple[str, ...],
    cftime_converter: CftimeConverter,
) -> xr.Dataset:
    """
    Convert some co-ordinates representing time into a time axis

    Parameters
    ----------
    inp
        Data to convert

    time_coords
        Co-ordinates from which to create the time axis

    cftime_converter
        Callable that converts the stacked time co-ordinates to
        :obj:`cftime.datetime`

    Returns
    -------
        Data with time axis
    """
    inp = inp.stack(time=time_coords)
    times = inp["time"].to_numpy()

    inp = inp.drop_vars(("time", *time_coords)).assign_coords(
        {"time": [cftime_converter(*t) for t in times]}
    )

    return inp


def split_time_to_year_month(
    inp: xr.Dataset,
    time_axis: str = "time",
) -> xr.Dataset:
    """
    Convert the time dimension to year and month without stacking

    This means there is still a single time dimension in the output, but there
    is now also accompanying year and month information

    Parameters
    ----------
    inp
        Data to convert

    Returns
    -------
        Data with year and month information for the time axis

    Raises
    ------
    NonUniqueYearMonths
        The years and months are not unique
    """
    out = inp.assign_coords(
        {
            "month": inp[time_axis].dt.month,
            "year": inp[time_axis].dt.year,
        }
    ).set_index({time_axis: ("year", "month")})

    # Could be updated when https://github.com/pydata/xarray/issues/7104 is
    # closed
    unique_vals, counts = np.unique(out[time_axis].values, return_counts=True)

    if (counts > 1).any():
        raise NonUniqueYearMonths(unique_vals, counts)

    return out


def convert_time_to_year_month(
    inp: xr.Dataset,
    time_axis: str = "time",
) -> xr.Dataset:
    """
    Convert the time dimension to year and month co-ordinates

    Parameters
    ----------
    inp
        Data to convert

    Returns
    -------
        Data with year and month co-ordinates
    """
    return split_time_to_year_month(
        inp=inp,
        time_axis=time_axis,
    ).unstack(time_axis)


def get_start_of_next_month(y: int, m: int) -> cftime.datetime:
    """
    Get start of next month

    Parameters
    ----------
    y
        Year

    m
        Month

    Returns
    -------
        Start of next month
    """
    if m == MONTHS_PER_YEAR:
        m_out = 1
        y_out = y + 1
    else:
        m_out = m + 1
        y_out = y

    # This may need to be refactored to allow the cftime_converter to be
    # injected, same idea as `convert_to_time`
    return cftime.datetime(y_out, m_out, 1)



"""
Exceptions
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Hashable

    import xarray as xr
    from pint.facets.plain.unit import PlainUnit


class CoordinateError(ValueError):
    """
    Raised when co-ordinates are not as expected

    Named "CoordinateError" to avoid confusion with
    :class:`pint.errors.DimensionalityError`
    """

    def __init__(
        self,
        exp_dims: tuple[Hashable, ...],
        found_dims: tuple[Hashable, ...],
        array: xr.DataArray,
        extras_ok: bool,
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        exp_dims
            Expected dimensions

        found_dims
            Actual dimensions

        array
            The array we're looking at

        extras_ok
            Could ``exp_dims`` be a subset of ``found_dims``, or did they
            have to be equal?
        """
        if extras_ok:
            error_msg = (
                f"Expected dimensions: `{exp_dims}`. These are not a subset "
                f"of the found dimensions: `{found_dims}`."
            )
        else:
            error_msg = (
                f"Expected dimensions: `{exp_dims}`. These are not equal "
                f"to the found dimensions: `{found_dims}`."
            )

        super().__init__(error_msg, array)


class NotPintQuantifiedError(ValueError):
    """
    Raised when xarray object hasn't been quantified with pint
    """

    def __init__(
        self,
        da: xr.DataArray,
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        da
            Unquantified :obj:`xr.DataArray`
        """
        error_msg = (
            f"{da.name} must have been quantified with pint using e.g. "
            "``.pint.quantify()``"
        )

        super().__init__(error_msg, da)


class DatasetIncompatibleUnitsError(ValueError):
    """
    Raised when an :obj:`xr.Dataset` object's contains units that incompatible

    Here incompatible means they are not of the same dimensionality i.e.
    cannot be converted into one another. Obviously this requirement is only
    applied in some cases
    """

    def __init__(
        self,
        base_unit: PlainUnit,
        incompatible_unit: PlainUnit,
        ds: xr.Dataset,
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        base_unit
            Base unit that was used in checking

        incompatible_unit
            Unit found to be incompatible during checking

        ds
            The dataset we're looking at
        """
        error_msg = f"{base_unit} is not compatible with {incompatible_unit}."

        super().__init__(error_msg, ds)



"""
Metdata options

This module could be refactored so that it is autogenerated from some other
source (or simply be an adapter to some other API) in future to remove the
need for manual updates.
"""
from __future__ import annotations

import re

ACTIVITY_ID_OPTIONS: tuple[str, ...] = ("input4MIPs",)
"""
Valid options for "activity_id""
"""

CONVENTION_OPTIONS: tuple[str, ...] = ("CF-1.7 CMIP-6.2", "CF-1.6")
"""
Valid options for "Convention""
"""

DATASET_CATEGORY_OPTIONS: tuple[str, ...] = (
    "aerosolProperties",
    "atmosphericState",
    "emissions",
    "GHGConcentrations",
    "landState",
    "ozone",
    "radiation",
    "SSTsAndSeaIce",
    "solar",
    "surfaceAir",
    "surfaceFluxes",
)
"""
Valid options for "dataset_category""
"""

INCLUDES_EMAIL_REGEX: re.Pattern[str] = re.compile(r"^.*?(\S+@\S+\.\S+).*$")
"""
Regular expression that checks there is something like an email somewhere

This is very loose and just provides a basic check to really avoid obvious
typos. It turns out writing a perfect regexp for email addresses is hard (see
e.g. https://stackoverflow.com/questions/201323/how-can-i-validate-an-email-address-using-a-regular-expression)
"""

CREATION_DATE_REGEX: re.Pattern[str] = re.compile(
    r"^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}Z$"
)
"""
Regular expression that checks the creation date is formatted correctly
"""

UUID_REGEX: re.Pattern[str] = re.compile(
    r"^hdl:21.14100\/[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}$"
)
"""
Regular expression that checks the creation date is formatted correctly
"""



"""
Testing related functions
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def get_call_kwargs(
    paras: tuple[tuple[str, Any]],
    extra_kwargs: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """
    Get kwargs used to call a callable.

    Useful when you want to test the default values of a function

    Parameters
    ----------
    paras
        Input parameters to test. Each element is the parameter's name and its
        value. If the value is ``None``, it is not included in the output.

    extra_kwargs
        Extra keyword arguments to the callable

    Returns
    -------
        Dictionary that can be used as the keyword arguments

    Examples
    --------
    >>> get_call_kwargs((("a", 3), ("b", None)))
    {'a': 3}
    >>> get_call_kwargs((("a", 3), ("b", "hello")))
    {'a': 3, 'b': 'hello'}
    >>> get_call_kwargs((("a", None), ("b", (1, 2))), extra_kwargs={"extra": 45})
    {'b': (1, 2), 'extra': 45}
    >>> # You have to be a bit careful with extra_kwargs because it can override things
    >>> get_call_kwargs((("a", None), ("b", (1, 2))), extra_kwargs={"b": 3})
    {'b': 3}
    """
    call_kwargs = {}

    for name, para in paras:
        if para is not None:
            call_kwargs[name] = para

    if extra_kwargs is not None:
        call_kwargs = call_kwargs | extra_kwargs

    return call_kwargs



"""
Tools for helping with :mod:`xarray`
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from carpet_concentrations.attrs_utils import (
    make_attrs_validator_compatible_single_input,
)
from carpet_concentrations.exceptions import (
    CoordinateError,
    DatasetIncompatibleUnitsError,
)

if TYPE_CHECKING:
    from collections.abc import Hashable


def check_dimensions(
    inp: xr.DataArray, exp_dims: tuple[Hashable, ...], extras_ok: bool = False
) -> None:
    """
    Check dimensions of an :obj:`xr.DataArray`

    Parameters
    ----------
    inp
        Data to check

    exp_dims
        Dimensions we expect the data to have

    extras_ok
        Is it ok if there are other dimensions present in the data,
        beyond those specified in ``extra_dims``?

    Raises
    ------
    CoordinateError
        Dimensions aren't as expected
    """
    dims = inp.dims

    if extras_ok:
        problem = not set(exp_dims).issubset(set(dims))
    else:
        problem = set(exp_dims) != set(dims)

    if problem:
        raise CoordinateError(exp_dims, dims, inp, extras_ok)


def calculate_weighted_area_mean_latitude_only(
    inp: xr.Dataset,
    variables: list[str],
    bounds_dim_name: str = "bounds",
    lat_name: str = "lat",
    lat_bounds_name: str = "lat_bounds",
) -> xr.Dataset:
    """
    Calculate area mean based on only latitude information

    :footcite:t:`kelly_savric_2020_computation`
    (https://doi.org/10.1111/tgis.12636)

    Parameters
    ----------
    inp
        :obj:`xr.Dataset` to process

    variables
        Variables of which to calculate the area-mean

    bounds_dim_name
        Name of the dimension which defines bounds

    lat_name
        Name of the latitude dimension

    lat_bounds_name
        Name of the latitude bounds variable

    Returns
    -------
        :obj:`xr.Dataset` with area-weighted mean of ``variables``
    """
    lat_bnds = inp[lat_bounds_name].pint.to("radian")

    # as they are constants, r^2 and longitude factors drop out
    # in area weights. Would have to be more careful with
    # longitudes if on a non-uniform grid.
    area_weighting = np.sin(lat_bnds).diff(dim=bounds_dim_name).squeeze()

    area_weighted_mean = (inp[variables] * area_weighting).sum(
        lat_name
    ) / area_weighting.sum(lat_name)

    # #8: allow dependency injection here
    keys_to_check = list(inp.data_vars.keys()) + list(inp.coords.keys())
    other_stuff = [v for v in keys_to_check if v not in variables]
    out = xr.merge([area_weighted_mean, inp[other_stuff]])

    return out


def check_all_units_compatible(ds: xr.Dataset) -> None:
    """
    Check all units in an :obj:`xr.Dataset` are compatible

    This only checks the data variables, not units of co-ordinates or
    dimensions.

    Here compatible means the units can be converted into one another

    Parameters
    ----------
    ds
        Dataset to check

    Raises
    ------
    DatasetIncompatibleUnitsError
        Some of the units in the dataset are incompatible
    """
    # #9: add dependency injection
    data_var_units = [
        v.data.units for v in ds.data_vars.values() if hasattr(v.data, "units")
    ]

    base_unit = data_var_units[0]
    for data_var_unit in data_var_units[1:]:
        if not data_var_unit.is_compatible_with(base_unit):
            raise DatasetIncompatibleUnitsError(base_unit, data_var_unit, ds)


check_all_units_compatible_attrs = make_attrs_validator_compatible_single_input(
    check_all_units_compatible
)



"""
Tools for helping with :mod:`xarray_pint`
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import pint

from carpet_concentrations.attrs_utils import (
    make_attrs_validator_compatible_single_input,
)
from carpet_concentrations.exceptions import NotPintQuantifiedError

if TYPE_CHECKING:
    import xarray as xr


def check_pint_quantified_array(inp: xr.DataArray) -> None:
    """
    Check that input can be converted using pint

    Parameters
    ----------
    inp
        Data to check

    Raises
    ------
    NotPintQuantifiedError
        ``inp`` is not pint compatible i.e. has not already been converted
        using pint via e.g. ``.pint.quantify()``
    """
    if not isinstance(inp.data, pint.Quantity):
        raise NotPintQuantifiedError(inp)


def check_pint_quantified_dataset(inp: xr.Dataset) -> None:
    """
    Check that all variables in an :obj:`xr.Dataset` can be converted using pint

    Parameters
    ----------
    inp
        Data to check
    """
    for val in inp.values():
        check_pint_quantified_array(val)


check_pint_quantified_array_attrs = make_attrs_validator_compatible_single_input(
    check_pint_quantified_array
)

check_pint_quantified_dataset_attrs = make_attrs_validator_compatible_single_input(
    check_pint_quantified_dataset
)



"""
Unit registry

Allows the user to pick a different registry if a different units system is
needed
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import openscm_units

if TYPE_CHECKING:
    import pint


UNIT_REGISTRY: pint.UnitRegistry = openscm_units.unit_registry
"""
Unit registry used throughout
"""
